#!/usr/bin/python
##############################################################################
#
# Copyright (c) 2002, 2006 Nexedi SARL and Contributors. All Rights Reserved.
#                    Bartlomiej Gorny <bartek@erp5.pl>
#
# WARNING: This program as such is intended to be used by professional
# programmers who take the whole responsability of assessing all potential
# consequences resulting from its eventual inadequacies and bugs
# End users who are looking for a ready-to-use solution with commercial
# garantees and support are strongly adviced to contract a Free Software
# Service Company
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
##############################################################################

import sys
from glob import glob
from xml.dom.minidom import parse

import config
sys.path.insert(0, config.uno_path)
import uno
from com.sun.star.beans import PropertyValue

class MimeMapper:

  # specify import filter for certain mimetypes.
  import_filter_by_mimetype = {
    'text/html':'HTML (StarWriter)'
    }

  # currently we use a unique extension for each export filter.
  extension_by_export_filter_dict = {
    'AportisDoc Palm DB': 'pdb',
    'BibTeX_Writer': 'bib',
    'Calc MS Excel 2007 XML': 'xlsx',
    'DIF': 'dif',
    'DocBook File': 'docbook.xml',
    'HTML (StarCalc)': 'calc.html',
    'HTML (StarWriter)': 'writer.html',
    'LaTeX_Writer': 'tex',
    'Impress MS PowerPoint 2007 XML': 'pptx',
    'Impress MS PowerPoint 2007 XML AutoPlay': 'ppsx',
    'Impress MS PowerPoint 2007 XML Template': 'potm',
    'MS Excel 2003 XML': '2003.xls.xml',
    'MS Excel 5.0/95': '5_0.xls',
    'MS Excel 5.0/95 Vorlage/Template': '5_0.xlt',
    'MS Excel 95': '95.xls',
    'MS Excel 95 Vorlage/Template': '95.xlt',
    'MS Excel 97': 'xls',
    'MS Excel 97 Vorlage/Template': 'xlt',
    'MS PowerPoint 97': 'ppt',
    'MS PowerPoint 97 AutoPlay': 'pps',
    'MS PowerPoint 97 Vorlage': 'pot',
    'MS WinWord 6.0': '6_0.doc',
    'MS Word 2003 XML': '2003.doc.xml',
    'MS Word 2007 XML': 'docx',
    'MS Word 95': '95.doc',
    'MS Word 97': 'doc',
    'MediaWiki': 'wiki.txt',
    'OpenDocument Drawing Flat XML': 'fodg',
    'OpenDocument Presentation Flat XML': 'fodp',
    'OpenDocument Spreadsheet Flat XML': 'fods',
    'OpenDocument Text Flat XML': 'fodt',
    'Pocket Excel': 'pxl',
    'PocketWord File': 'psw',
    'Rich Text Format': 'rtf',
    'SYLK': 'slk',
    'StarCalc 3.0': '3_0.sdc',
    'StarCalc 3.0 Vorlage/Template': '3_0.sdc.vor',
    'StarCalc 4.0': '4_0.sdc',
    'StarCalc 4.0 Vorlage/Template': '4_0.sdc.vor',
    'StarCalc 5.0': '5_0.sdc',
    'StarCalc 5.0 Vorlage/Template': 'sdc.vor',
    'StarDraw 3.0': '3_0.sda',
    'StarDraw 3.0 (StarImpress)': 'odp.3_0.sda',
    'StarDraw 3.0 Vorlage (StarImpress)': 'odp.3_0.sda.vor',
    'StarDraw 3.0 Vorlage': '3_0.sda.vor',
    'StarDraw 5.0': 'sda',
    'StarDraw 5.0 (StarImpress)': 'odp.sda',
    'StarDraw 5.0 Vorlage': 'sda.vor',
    'StarDraw 5.0 Vorlage (StarImpress)': 'odp.sda.vor',
    'StarImpress 4.0': '4_0.sdd',
    'StarImpress 4.0 Vorlage': '4_0.sdd.vor',
    'StarImpress 5.0': 'sdd',
    'StarImpress 5.0 Vorlage': 'sdd.vor',
    'StarOffice XML (Calc)': 'sxc',
    'StarOffice XML (Draw)': 'sxd',
    'StarOffice XML (Impress)': 'sxi',
    'StarOffice XML (Writer)': 'sxw',
    'StarWriter 3.0': '3_0.sdw',
    'StarWriter 3.0 Vorlage/Template': '3_0.sdw.vor',
    'StarWriter 4.0': '4_0.sdw',
    'StarWriter 4.0 Vorlage/Template': '4_0.sdw.vor',
    'StarWriter 5.0': 'sdw',
    'StarWriter 5.0 Vorlage/Template': 'sdw.vor',
    'Text': 'txt',
    'Text (encoded)': 'enc.txt',
    'Text - txt - csv (StarCalc)': 'csv',
    'UOF presentation': 'uop',
    'UOF spreadsheet': 'uos',
    'UOF text': 'uot',
    'XHTML Calc File': 'calc.xhtml',
    'XHTML Draw File': 'draw.xhtml',
    'XHTML Impress File': 'impr.xhtml',
    'XHTML Writer File': 'writer.xhtml',
    'calc8': 'ods',
    'calc8_template': 'ots',
    'calc_StarOffice_XML_Calc_Template': 'stc',
    'calc_pdf_Export': 'calc.pdf',
    'dBase': 'dbf',
    'draw8': 'odg',
    'draw8_template': 'otg',
    'draw_StarOffice_XML_Draw_Template': 'std',
    'draw_bmp_Export': 'bmp',
    'draw_emf_Export': 'emf',
    'draw_eps_Export': 'draw.eps',
    'draw_flash_Export': 'draw.swf',
    'draw_gif_Export': 'gif',
    'draw_html_Export': 'draw.html',
    'draw_jpg_Export': 'jpg',
    'draw_met_Export': 'met',
    'draw_pbm_Export': 'pbm',
    'draw_pct_Export': 'pct',
    'draw_pdf_Export': 'draw.pdf',
    'draw_pgm_Export': 'pgm',
    'draw_png_Export': 'png',
    'draw_ppm_Export': 'ppm',
    'draw_ras_Export': 'ras',
    'draw_svg_Export': 'svg',
    'draw_svm_Export': 'svm',
    'draw_tif_Export': 'tiff',
    'draw_wmf_Export': 'wmf',
    'draw_xpm_Export': 'xpm',
    'impress8': 'odp',
    'impress8_draw': 'odp.odg',
    'impress8_template': 'otp',
    'impress_StarOffice_XML_Draw': 'odp.sxd',
    'impress_StarOffice_XML_Impress_Template': 'sti',
    'impress_bmp_Export': 'impr.bmp',
    'impress_emf_Export': 'impr.emf',
    'impress_eps_Export': 'impr.eps',
    'impress_flash_Export': 'impr.swf',
    'impress_gif_Export': 'impr.gif',
    'impress_html_Export': 'impr.html',
    'impress_jpg_Export': 'impr.jpg',
    'impress_met_Export': 'impr.met',
    'impress_pbm_Export': 'impr.pbm',
    'impress_pct_Export': 'impr.pct',
    'impress_pdf_Export': 'impr.pdf',
    'impress_pgm_Export': 'impr.pgm',
    'impress_png_Export': 'impr.png',
    'impress_ppm_Export': 'impr.ppm',
    'impress_ras_Export': 'impr.ras',
    'impress_svg_Export': 'impr.svg',
    'impress_svm_Export': 'impr.svm',
    'impress_tif_Export': 'impr.tiff',
    'impress_wmf_Export': 'impr.wmf',
    'impress_xpm_Export': 'impr.xpm',
    'placeware_Export': 'pwp',
    'writer8': 'odt',
    'writer8_template': 'ott',
    'writer_StarOffice_XML_Writer_Template': 'stw',
    'writer_pdf_Export': 'pdf'
    }

  document_type_list = ['text', 'spreadsheet', 'presentation', 'drawing']
  odf_type_list = ['odt', 'ods', 'odp', 'odg']
  odf_filter_list = ['writer8', 'calc8', 'impress8', 'draw8']

  def __init__(self):
    self.collectFilterInfo()
    self.setFilterInfo()
    self.filterDocumentType()

  def collectFilterInfo(self):
    """
    collect Filter information and Type information from OOo's directory.
    """
    def getFileList(dir):
      return glob('%s/*' % dir)

    filter_dir = '%s/../share/registry/modules/org/openoffice/TypeDetection/Filter' % config.uno_path
    type_dir = '%s/../share/registry/modules/org/openoffice/TypeDetection/Types' % config.uno_path
    filter_file_list = getFileList(filter_dir)
    type_file_list = getFileList(type_dir)
    if not filter_file_list or not type_file_list:
      raise Exception('no OOo filter found.')

    def prop_by_name(node, name):
      for x in node.getElementsByTagName('prop'):
        if x.getAttribute('oor:name') == name:
          try:
            value = x.getElementsByTagName('value')[0]
            return value.firstChild.data.encode('utf8')
          except IndexError:
            return None
      return None

    def node_by_name(node, name):
      for x in node.getElementsByTagName('node'):
        if x.getAttribute('oor:name') == name:
          return x
      return None

    def sort_index(flag_list, extension):
      num = 0
      if 'OWN' not in flag_list:
        num += 10000
      if 'ALIEN' in flag_list:
        num += 1000
      if '3RDPARTYFILTER' in flag_list:
        num += 100
      elif 'SUPPORTSSELECTION' in flag_list:
        num += 200
      if 'PREFERRED' not in flag_list:
        num += 10
      if 'TEMPLATEPATH' in flag_list:
        num += 1
      return '%05.f.%s' % (num, extension)

    type_list = []
    for x in type_file_list:
      type_list.extend(
        node_by_name(parse(x), 'Types').getElementsByTagName('node'))

    filter_list = []
    for x in filter_file_list:
      filter_list.extend(
        node_by_name(parse(x), 'Filters').getElementsByTagName('node'))

    document_type_dict = {
      'com.sun.star.text.TextDocument':'text',
      'com.sun.star.sheet.SpreadsheetDocument':'spreadsheet',
      'com.sun.star.drawing.DrawingDocument':'drawing',
      'com.sun.star.presentation.PresentationDocument':'presentation',
      }

    self.document_type_by_extension_dict = {}
    self.document_type_by_mimetype_dict = {}
    self.document_type_by_input_filter_dict = {}
    self.extension_by_mimetype_dict = {}

    for type_node in type_list:
      preferred_filter = prop_by_name(type_node, 'PreferredFilter')
      if preferred_filter is None:
        continue
      filter_node = [f for f in filter_list if f.getAttribute('oor:name') == preferred_filter][0]
      document_type = document_type_dict.get(prop_by_name(filter_node,
                                                          'DocumentService'))
      if document_type is None:
        continue
      flags = prop_by_name(filter_node, 'Flags')
      if flags.find('IMPORT') == -1:
        continue
      # we found a entry split by ';' instead of ' '.
      extensions = prop_by_name(type_node, 'Extensions').replace(';', ' ').split(' ')
      if extensions != '':
        for extension in extensions:
          self.document_type_by_extension_dict[extension] = document_type
      mimetype = prop_by_name(type_node, 'MediaType')
      if mimetype is not None:
        self.document_type_by_mimetype_dict[mimetype] = document_type
      input_filter = filter_node.getAttribute('oor:name').encode('utf8')
      self.document_type_by_input_filter_dict[input_filter] = document_type
      self.extension_by_mimetype_dict[mimetype] = extensions[0]

    # we don't know why but HTML filter is missing in *.xcu
    self.document_type_by_extension_dict['html'] = 'text'
    self.document_type_by_extension_dict['htm'] = 'text'
    self.document_type_by_mimetype_dict['text/html'] = 'text'
    self.document_type_by_input_filter_dict['HTML (StarWriter)'] = 'text'
    self.extension_by_mimetype_dict['text/html'] = 'html'

    # we want more meaningful name for several filters
    ui_name_dict = {
      'Calc 8':'ODF Spreadsheet',
      'Calc 8 Template':'ODF Spreadsheet Template',
      'Calc 6.0':'OpenOffice.org 1.0 Spreadsheet',
      'Calc 6.0 Template':'OpenOffice.org 1.0 Spreadsheet Template',
      'Draw 8':'ODF Drawing',
      'Draw 8 Template':'ODF Drawing Template',
      'StarOffice XML (Draw)':'OpenOffice.org 1.0 Drawing',
      'Draw 6.0 Template':'OpenOffice.org 1.0 Drawing Template',
      'Impress 8':'ODF Presentation',
      'Impress 8 Template':'ODF Presentation Template',
      'StarOffice XML (Impress)':'OpenOffice.org 1.0 Presentation',
      'Impress 6.0 Template':'OpenOffice.org 1.0 Presentation Template',
      'Writer 8':'ODF Text Document',
      'Writer 8 Template':'ODF Text Document Template',
      'StarOffice XML (Writer)':'OpenOffice.org 1.0 Text Document',
      'Writer 6.0 Template':'OpenOffice.org 1.0 Text Document Template',
      'Text - txt - csv (StarOffice Calc)':'Text CSV',
      'impress_StarOffice_XML_Draw':'OpenOffice.org 1.0 Drawing',
      }

    self.filter_property_dict = {}

    for filter_node in filter_list:
      document_type = document_type_dict.get(prop_by_name(filter_node,
                                                          'DocumentService'))
      if document_type is None:
        continue
      export_filter = filter_node.getAttribute('oor:name').encode('utf8')
      flag_list = prop_by_name(filter_node, 'Flags').split(' ')
      if 'EXPORT' not in flag_list or 'TEMPLATEPATH' in flag_list or \
          'STARTPRESENTATION' in flag_list:
        continue
      type_name = prop_by_name(filter_node, 'Type')
      try:
        type_node = [t for t in type_list if t.getAttribute('oor:name') == type_name][0]
      except IndexError:
        continue
      ui_name = prop_by_name(type_node, 'UIName')
      if ui_name is None:
        continue
      if ui_name.find('%') != -1:
        ui_name = filter_node.getAttribute('oor:name').encode('utf8')
      media_type = prop_by_name(type_node, 'MediaType') or \
                   'application/octet-stream'
      # we found a entry split by ';' instead of ' '.
      extension = prop_by_name(type_node, 'Extensions').replace(';', ' ').split(' ')[0]
      self.filter_property_dict[export_filter] = {
        'document_type':document_type,
        'extension':extension,
        'mime_type':media_type,
        'label':ui_name_dict.get(ui_name, ui_name),
        'sort_index':sort_index(flag_list, extension),
        }

    self.export_filter_by_extension_dict = {}
    for k,v in self.extension_by_export_filter_dict.items():
      self.export_filter_by_extension_dict[v] = k

    self.document_type_dict = {}
    for document_type in self.document_type_list:
      tmp_list = self.filter_property_dict.items()
      tmp_list.sort(lambda a,b:cmp(a[1]['sort_index'], b[1]['sort_index']))
      self.document_type_dict[document_type] = \
        [self.extension_by_export_filter_dict[k] for k, v in tmp_list \
         if v['document_type'] == document_type and k in self.extension_by_export_filter_dict.keys()]

  def setFilterInfo(self):
    # specify more properties for several filters
    if 'Text (encoded)' in self.filter_property_dict:
      self.filter_property_dict['Text (encoded)']['export_props'] = ( \
        PropertyValue('FilterFlags', 0, 'UTF8,LF', 0),
        )
    if 'draw_html_Export' in self.filter_property_dict:
      self.filter_property_dict['draw_html_Export']['export_props'] =  ( \
      PropertyValue('FilterData',0,uno.Any(
        '[]com.sun.star.beans.PropertyValue',
        (PropertyValue('Format', 0, 2, 0),),), 0), # PNG format
      )
    if 'impress_html_Export' in self.filter_property_dict:
      self.filter_property_dict['impress_html_Export']['export_props'] = ( \
      PropertyValue('FilterData',0,uno.Any(
        '[]com.sun.star.beans.PropertyValue',
        (PropertyValue('IsExportNotes', 0, True, 0),
         PropertyValue('Format', 0, 2, 0),),), 0), # PNG format
      )
    if 'impress_pdf_Export' in self.filter_property_dict:
      self.filter_property_dict['impress_pdf_Export']['export_props'] = ( \
      PropertyValue('FilterData',0,uno.Any(
        '[]com.sun.star.beans.PropertyValue',
        (PropertyValue('ExportNotesPages', 0, True, 0),),), 0),
      )

  def filterDocumentType(self):
    """
      This method is run upon init; it filters entries in document_type_dict (the dictionary which is used to build
      allowed_target_item_list) according to what is specified in oood.conf file,
      so that the system administrator can limit the list of target format to what is really
      needed.
    """
    conf = config.config
    for document_type in self.document_type_dict.keys():
      section = document_type.title()
      use_formats = conf.has_option(section, 'use_formats') and conf.get(section, 'use_formats') or 'All'
      if use_formats.lower() == 'all':
        use_formats = None
      else:
        use_formats = [s.strip() for s in use_formats.split(',')]
      skip_formats = conf.has_option(section, 'skip_formats') and conf.get(section, 'skip_formats') or 'None'
      if skip_formats.lower() == 'all':
        skip_formats = self.document_type_dict[document_type]
      else:
        skip_formats = [s.strip() for s in skip_formats.split(',')]
      def useFormat(format):
        if use_formats:
          return format in use_formats
        if skip_formats:
          return format not in skip_formats
        return True
      self.document_type_dict[document_type] = [format for format in self.document_type_dict[document_type] if useFormat(format)]

  def getAllowedTo(self, mimetype=None, document_type=None, extension=None):
    """
      returns a list of tuples made of (type, label)
      for all formats to which we can convert the given mimetype
      We return all formats of a given type, except for the type of original document if known
    """
    if document_type is not None:
      pass
    elif mimetype is not None:
      document_type = self._mimeToType(mimetype)
    elif extension is not None:
      document_type = self._extToType(extension)
    if not document_type:
      return []
    return [(ext,
             self.getTypePropertyDict(extension=ext)['label']) \
            for ext in self.document_type_dict[document_type]]

  def _mimeToType(self, mimetype):
    """
      Helper function to map mimetype to document type
    """
    return self.document_type_by_mimetype_dict.get(mimetype, None)

  def _extToType(self, extension):
    """
      Helper function to map filename extension to document type
    """
    return self.document_type_by_extension_dict.get(extension, None)

  def _filterNameToType(self, filter_name):
    """
      Helper function to map filter name to document type
    """
    return self.document_type_by_input_filter_dict.get(filter_name, None)

  def getMimeFor(self, format):
    """
      return mime type for a given format (format is file name extension)
    """
    try:
      export_filter = self.export_filter_by_extension_dict[format]
      return self.filter_property_dict[export_filter]['mime_type']
    except KeyError:
      raise Exception('no mime type for format ' + format)

  def isOdf(self, extension=None, filter_name=None):
    """
      Method for a quick check if a document is ODF
    """
    if filter_name is not None:
      return filter_name in self.odf_filter_list
    if extension is not None:
      return extension in self.odf_type_list

  def getOdfInfo(self, filter_name=None):
    """
      Get info about ODF format appropriate for the given format
    """
    if filter_name is not None:
      document_type = self._filterNameToType(filter_name)
    if not document_type:
      return
    odf_filter = [x for x in self.odf_filter_list \
                  if self.filter_property_dict[x]['document_type'] == document_type][0]
    odf_filter_property = self.filter_property_dict[odf_filter]
    # XXX it can be
    # extension = odf_filter_property['extension']
    extension = self.extension_by_export_filter_dict[odf_filter]
    return extension, odf_filter_property['mime_type']

  def getTypePropertyDict(self, extension=None, filter_name=None):
    """
      Get properties for generating file in a given format
      (format is given as mimetype or extension)
    """
    if filter_name is None:
      filter_name = self.getFilterForExt(extension)
    return self.filter_property_dict.get(filter_name, None)

  def getExtForMime(self, mimetype):
    return self.extension_by_mimetype_dict.get(mimetype, None)

  def getFilterForExt(self, format):
    return self.export_filter_by_extension_dict.get(format, None)

  def getInputFilterForMime(self, mimetype):
    return self.import_filter_by_mimetype.get(mimetype, None)

mimemapper = MimeMapper()

if __name__=='__main__':
  import pprint
  pp = pprint.PrettyPrinter(indent=2)
  pp.pprint(mimemapper.getAllowedTo(mimetype='text/html'))
  pp.pprint(mimemapper.getAllowedTo(mimetype='application/x-shockwave-flash'))
  pp.pprint('========================================')
  for k,v in mimemapper.document_type_dict.items():
    pp.pprint(k)
    pp.pprint(v)
    pp.pprint('========================================')
  pp.pprint(mimemapper.isOdf(extension='sxw'))
  pp.pprint(mimemapper.isOdf(extension='odp'))
  pp.pprint(mimemapper.isOdf(filter_name='calc8'))
  pp.pprint('========================================')
  pp.pprint(mimemapper.getOdfInfo(filter_name='writer8'))
  pp.pprint('========================================')
  pp.pprint(mimemapper.getAllowedTo(mimetype='application/vnd.oasis.opendocument.text'))
  pp.pprint(mimemapper.getAllowedTo(extension='rtf'))
  pp.pprint(mimemapper.getAllowedTo(document_type='text'))
