#!/usr/bin/python
##############################################################################
#
# Copyright (c) 2002, 2006 Nexedi SARL and Contributors. All Rights Reserved.
#                    Bartlomiej Gorny <bartek@erp5.pl>
#                    Kevin Deldycke <kevin@nexedi.com>
#                    Jacek Medrzycki <jacek@erp5.pl>
#
# WARNING: This program as such is intended to be used by professional
# programmers who take the whole responsability of assessing all potential
# consequences resulting from its eventual inadequacies and bugs
# End users who are looking for a ready-to-use solution with commercial
# garantees and support are strongly adviced to contract a Free Software
# Service Company
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
##############################################################################

import getopt
import sys
import os
import errno
import signal
import atexit
import base64
from logger import Log

# URE_BOOTSTRAP environment variable is required for OOo-3.0
import config
fundamentalrc_file = '%s/fundamentalrc' % config.prog_dir
if os.path.exists(fundamentalrc_file) and \
   not os.environ.has_key('URE_BOOTSTRAP'):
  os.environ['URE_BOOTSTRAP'] = 'vnd.sun.star.pathname:%s' % fundamentalrc_file

def showServerStatus():
  """
    Show Server details and statistics.
  """
  # TODO: must be done (!)
  print "===== Server Statistics ====="
  print "Status: %s" % "TODO"
  print "Uptime: %s" % "TODO"
  print "Auto-reboot since launch: %s" % "TODO"


def usage():
  print """Usage: %s [options]
  --start
      Start the server.
  --stop
      Stop the server.
  --offset=n
      With --start specifies the port offset (negative numers are allowed)
      for the server to be started at
  --multiserv=range
      Starts a range of servers, each with different 
      port offset. Range can be:
      m:n where m is start port and m is end port offset, or
      k which means range 0..k-1
  --status, --stat
      Show server statistics and status.
  -d, --debug
      Start the server in debug mode.
  -h, --help
      Show this screen.
""" % sys.argv[0]

lock_file_path=''

def at_exit_handler(lock_file_path):
  Log.info("Server - Remove lock file and stop server")
  print "Server - Remove lock file and stop server"
  if os.path.exists(lock_file_path):
    os.remove(lock_file_path)
  os._exit(0)

def sig_term_handler(signum, frame):
  sys.exit(0)

def startMultiServer(param):

  import config
  from threading import Thread
  import socket
  from time import sleep

  start=0
  end = None

  params = param.split(':')
  if len(params)==1:
    end = int(params[0])
  else:
    start = int(params[0])
    end = int(params[1])+1

  server_range = range(start, end)

  MAX_RETRYS=15

  # Create a lock file to not run more than one server and to save the pid
  lock_file_path = os.path.join(config.run_dir, 'server%s_pid.lock'%offset_str)
  if os.path.exists(lock_file_path):
    message = "Server is already running (lock file %s exists)" % \
                                           lock_file_path
    Log.info(message)
    print message
    sys.exit(1)

  class Starter(Thread):
    def run(self):
      os.system('./runserw.py --start --offset=%s' % self.serv_num)
  print server_range

  server_threads = []
  for serv_num in server_range:
    #try to start a server
    starter=Starter()
    starter.serv_num = serv_num
    starter.start()
    print "starting server# %s" % serv_num

    #wait until it is possible to connect to server's port
    rn = 0
    while rn < MAX_RETRYS:
      try:
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.connect((config.server_host, config.server_port + serv_num))
        s.close()
        break
      except socket.error:
        print "waiting for server# %s to respond" % serv_num
        pass

      rn+=1
      sleep(5)
      server_threads.append(starter)
  
    if rn == MAX_RETRYS: 
      raise Exception, "Server# %s is not responding" % serv_num
      sys.exit(1)

    print "Server# %s started" % serv_num

  # XXX Ugly but working. Perhaps better solution to find in the future
  # After starting a server, previous servers loose (for unknown reason) contact with its ooo instance.
  # The first call then often fail, and then server reloads ooo and starts to work ok.
  # So let's make each server converting simple text document just to give it chance to reload
  print 'kicking servers'
  import xmlrpclib
  for serv_num in server_range:
    port = config.server_port + serv_num 
    sp=xmlrpclib.ServerProxy('http://%s:%s' % (config.server_host, port))
    enc_data = base64.encodestring('simple text\n')
    response = sp.generate({'data':enc_data,'extension':'odt'})
    #we are not interseted in returned value anyway ;)

  print 'All servers should be operational now'
  atexit.register(at_exit_handler, lock_file_path)
  signal.signal(signal.SIGTERM, sig_term_handler)

  lock_file = open(lock_file_path, 'w')
  server_pid = os.getpid()
  lock_file.write(str(server_pid))
  lock_file.close()
  message = "Multi Server - Started as process %s" % server_pid
  Log.info(message)
  print message

  #wait for children thread
  for t in server_threads:
    t.join()

  print 'quniec'

if __name__ == '__main__':
  # Parse options and parameters
  try:
    opts, args = getopt.getopt( sys.argv[1:]
           , "hd"
           , ["help", "stop", "status", "stat", "start", "debug", "offset=", "multiserv="]
           )
    # Check args number
    if len(opts) == 0:
      raise getopt.GetoptError('', '')
  except getopt.GetoptError:
    # Print help information and exit on command line error
    usage()
    sys.exit(2)

  # Initialize default variables
  debug = False
  start = False
  stop  = False
  offset = 0
  offset_str = ''
  multi = False

  # Start action according to parameters
  for o, a in opts:
    if o in ("-h", "--help"):
      usage()
      sys.exit(0)
    elif o in ("--status", "--stat"):
      showServerStatus()
      sys.exit(0)
    elif o in ("-d", "--debug"):
      debug = True
    elif o in ("--stop"):
      stop = True
    elif o in ("--start"):
      start = True
    elif o in ("--offset"):
      try:
        offset = int(a)
      except ValueError:
        usage()
        sys.exit(1)
      offset_str = str(offset) #this is to normalize offset parameters: 01 should be the same as 1
    elif o in ("--multiserv"):
      multi_param = a
      multi = True

  if (start or debug) and multi:
    print "start or debug and multiserv are mutually exclusive"
    usage()
    sys.exit(1)

  if stop:
    #########################################################
    ## Use the lock/pid file to terminate the server.
    #########################################################
    import config
    import glob
    lock_file_path = os.path.join(config.run_dir, 'server_pid.lock')
    # Get the server process id which is written in the lock file.
    if not os.path.exists(lock_file_path):
      Log.info("Server - Can't stop server: it was not running !")
      sys.exit(1)

    #Perhaps we are running multiserver? try to get list of lock files for children
    lock_file_path_list = glob.glob(os.path.join(config.run_dir, 'server?*_pid.lock'))
    #add master lock file is the list is still empty
    #there is no need to add master lock file if the list is not empty
    #(that is, if there are children servers), because parent will die
    #after all of its children die (very sad indeed)
    if not lock_file_path_list:
      lock_file_path_list.append(lock_file_path)
    for lock_file_path in lock_file_path_list:
      lock_file = open(lock_file_path, 'r')
      server_pid = int(lock_file.read())
      lock_file.close()
      os.kill(server_pid, signal.SIGTERM)

    Log.info("Server - Stopped properly")
    sys.exit(0)


  elif start or debug:
    #########################################################
    ## Run the XML-RPC server
    #########################################################
    import config
    # check if we have tmp dir and it is writeable
    tmp_dir = config.tmp_dir+offset_str
    try:
      os.makedirs(tmp_dir)
    except OSError:
      pass
    if not os.access(tmp_dir, os.W_OK):
      message = 'ERROR: %s must exist and must be writeable' % tmp_dir
      Log.info(message)
      print message
      sys.exit(1)
    import start as starter
    starter.startVirtualFrameBuffer()
    from dispatcher import Dispatcher, MySerw

    if debug:
      Log.info("Server %s - Started in debug mode" % offset)

    # Create a lock file to not run more than one server and to save the pid
    lock_file_path = os.path.join(config.run_dir, 'server%s_pid.lock'%offset_str)
    if os.path.exists(lock_file_path):
      f = file(lock_file_path, 'r')
      try:
        pid = int(f.read())
      finally:
        f.close()
      try:
        os.kill(pid, 0)
      except OSError, err:
        if err[0] == errno.ESRCH:
          Log.info('Removing stale lock file %s' % lock_file_path)
          os.unlink(lock_file_path)
      else:
        message = "Server is already running (lock file %s exists)" % \
                                             lock_file_path
        Log.info(message)
        print message
        sys.exit(1)

    lock_file = open(lock_file_path, 'w')
    server_pid = os.getpid()
    lock_file.write(str(server_pid))
    lock_file.close()
    message = "Server - Started as process %s" % server_pid
    Log.info(message)
    print message
    
    atexit.register(at_exit_handler, lock_file_path)
    signal.signal(signal.SIGTERM, sig_term_handler)

    # Start the server core
    disp = Dispatcher(offset_str)

    # Some debug stuff
    if debug:
      d = open('test.odt').read()
      # XXX proc is not defined
      print proc.run_makepdf('test.odt', base64.encodestring(d))[0]
      #d=open('doc/test.doc').read()
      #print proc.run_setmetadata('test.odt',base64.encodestring(d),{'title':'zz'})[0]

    # Run the server forever
    server_port = config.server_port + offset
    ser = MySerw((config.server_host, server_port))
    ser.register_instance(disp)
    Log.info("Server - Will serve indefinitely until you kill or interrupt it...")
    # do not output to stderr
    sys.stderr = open('/dev/null', 'w')
    ser.serve_forever()

    sys.exit(0)
  elif multi:
    startMultiServer(multi_param)
  
  else:
    # Exit on general error
    usage()
    sys.exit(1)
